/*************************************************************************\
 *	eplot.c
 *	Photon graphics for 'euler' program.

 *	D. Vainapel
 *	davidv@elisra.com

 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the  Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 * USA.
\*************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#include <Ph.h>
#include <Pt.h>
#include <Ap.h>
#include "source/plot_msg.h"

// Euler coordinates range
#define XMAX  1000.0
#define YMAX  1000.0

#define MAX_PLOT 300

// memory buffer increments
#define BLOCK 0x10000

#define WinX( x ) ( x * mx  + marginX )
#define WinY( y ) ( y * my  + marginY )

PhRect_t	rect;
double		marginX, marginY;
int			numOfObjects = 0;
float		mx,my;		// Translation to pixel
char 		*img;
PtWidget_t	*plotw;
int waitForMouse = 0;

PgColor_t pColor[] = {
	Pg_WHITE,
	Pg_BLACK,
	PgRGB( 0, 168, 0 ),
	PgRGB( 0, 0, 168 ),
	PgRGB( 0, 0, 256 ),
	PgRGB( 168, 0, 0 ),
	PgRGB( 0, 168, 168 ),
	PgRGB( 168, 0, 168 ),
	PgRGB( 168, 84, 0 ),
	PgRGB( 84, 255, 84 ),
	PgRGB( 84, 255, 255 ),
	PgRGB( 255, 84, 84 ),
	PgRGB( 255, 84, 255 ),
	PgRGB( 255, 255, 84 ),
	PgRGB( 84, 84, 255 ),
	PgRGB( 168, 168, 168 ),
	PgRGB( 84, 84, 84 ),
};

PtWidget_t* CreateRawWidget( PtWidget_t * );
int ResizeCallback( PtWidget_t *, ApInfo_t *, PtCallbackInfo_t * );
int GetData( int, void *, unsigned );

int main()
{
    PtArg_t args[2];
    PtWidget_t *window;
    PtCallback_t resizecall;

	resizecall.event_f = &ResizeCallback;
    PtSetArg( &args[0], Pt_ARG_WINDOW_TITLE, "Euler", 0 );
    PtSetArg( &args[1], Pt_CB_RESIZE, &resizecall, 1); 
    if ((window = PtAppInit(NULL, 0, 0, 2, args)) == NULL) 
   	    return EXIT_FAILURE;
	plotw = CreateRawWidget( window);
    PtRealizeWidget(window); 
	// Nonblocking input
	fcntl( STDIN_FILENO, F_SETFL, O_NONBLOCK );
	PtAppAddFd( 0, STDIN_FILENO, Pt_FD_READ | Pt_FD_NOPOLL, GetData, 0);
    PtMainLoop(); 
    return EXIT_SUCCESS;
}

//----------------------------------------------------------------
// euler function 'graph'
//----------------------------------------------------------------
void DrawGraph( Graph *msg )
{
	PhPoint_t *p;
	PhPoint_t origin = {0,0};
	int i, m, n, inptr = 0;
	p = (PhPoint_t*) malloc( MAX_PLOT * sizeof(PhPoint_t) );
	if( p == 0 )
		return;
	n = msg->n;
	PgSetStrokeColor( pColor[ msg->color ] );
	while( n > 0 ) {
		m = min( n, MAX_PLOT );
		for( i = 0; i < m; i++ ) {
			p[i].x = WinX( msg->p[ inptr ++ ] );
			p[i].y = WinY( msg->p[ inptr ++ ] );
		}
		if( msg->connect )
			PgDrawPolygon( p, m, &origin, Pg_DRAW_STROKE );
		else
			PgDrawPixelArray( p, m, &origin );
		PgFlush();
		n -= m;
	}
	free( p );
}

//--------------------------------------------- 
// Color 1, 2, 3 for r, g, b else white
//----------------------------------------------
void SetHue( float hue , int color )
{
	int r = 255, g = 255, b = 255;
	switch( color ) {
	case 1:
		r = 0;	
		break;
	case 2:
		g = 0;	
		break;
	case 3:
		b = 0;	
		break;
	}
	PgSetFillColor( PgRGB( (int)(r * hue), (int)(b * hue), (int)(g *hue)) );
	PgSetStrokeColor( pColor[1] );
}

//----------------------------------------------------------------
// euler function 'line'
//----------------------------------------------------------------
void DrawLine( Gline *msg)
{
	int x, y, x1, y1;
	x = WinX( msg->c );
	y = WinY( msg->r );
	x1 = WinX( msg->c1 );
	y1 = WinY( msg->r1 );
	PgSetStrokeColor( pColor[ msg->color ] );
	PgDrawILine( x, y, x1, y1 );
}

//----------------------------------------------------------------
// euler function 'fill'
//----------------------------------------------------------------
void DrawFill( Gfill *msg )
{
	PhPoint_t *p;
	PhPoint_t origin = {0,0};
	int i;
	p = (PhPoint_t*) malloc( msg->n * sizeof(PhPoint_t) );
	for( i = 0; i < msg->n; i++ ) {
		p[i].x = WinX( msg->p[i * 2] );
		p[i].y = WinY( msg->p[i * 2+1] );
	}
	SetHue( msg->hue, msg->color );
	PgDrawPolygon( p, msg->n, &origin, Pg_DRAW_FILL_STROKE | Pg_CLOSED );
	free( p );
}

//----------------------------------------------------------------
// euler function 'text'
//----------------------------------------------------------------
static void DrawText( Gtext *msg)
{
	static int flag[3] = { Pg_TEXT_LEFT, Pg_TEXT_CENTER, Pg_TEXT_RIGHT }; 
	PhPoint_t p;
	PgSetTextColor( pColor[ msg->color ] );
	p.x = WinX( msg->c );
	p.y = WinY( msg->r );
	if( msg->r <= 1 )
		p.y -= marginY / 3;
	else if( msg->r > 999 ) 
		p.y += (marginY * 2) / 3 ;
	if( msg->c <= 0 ) {
		p.x = 5;
		PgDrawText( msg->text, strlen(msg->text), &p, Pg_TEXT_LEFT );
	} else 
		PgDrawText( msg->text, strlen(msg->text), &p, flag[msg->alignment] );
}

//----------------------------------------------------------------
// euler function 'marker'
//----------------------------------------------------------------
static void DrawMarker( Gmarker *msg)
{
	PhPoint_t radii={3,3}, center;
	PgSetStrokeColor( pColor[ msg->color ] );
	center.x = WinX( msg->c );
	center.y = WinY( msg->r );
	switch( msg->type )	{
	case 0:
		PgDrawILine( center.x-3, center.y-3, center.x+3, center.y+3 );
		PgDrawILine( center.x+3, center.y-3, center.x-3, center.y+3 );
		break;
	case 1:
		PgDrawEllipse( &center, &radii, Pg_DRAW_STROKE);
		break;
	case 2:
		PgDrawILine( center.x-3, center.y, center.x+3, center.y );
		PgDrawILine( center.x, center.y-3, center.x, center.y+3 );
		break;
	case 3:
		PgDrawILine( center.x-3, center.y, center.x+3, center.y );
		PgDrawILine( center.x, center.y-3, center.x, center.y+3 );
		break;
	default:
		PgDrawIRect( center.x-3, center.y-3, center.x+3, center.y+3,
			Pg_DRAW_STROKE );
		break;
	}
}

//----------------------------------------------------------------
// euler function 'bar'
//----------------------------------------------------------------
void DrawBar( Gbar *msg)
{
	int x,y,x1,y1;
	x = WinX( msg->c );
	y = WinY( msg->r );
	x1 = WinX( msg->c1 );
	y1 = WinY( msg->r1 );
	SetHue( msg->hue, msg->color );
	if( msg->border ) {
		PgSetStrokeColor( pColor[1] );
		PgDrawIRect( x, y, x1, y1, Pg_DRAW_FILL_STROKE );
	}
	else {
		PgDrawIRect( x, y, x1, y1, Pg_DRAW_FILL );
	}
}

//-----------------------------------------------------
// Select graphics function
//-----------------------------------------------------
static void DrawObjects()
{
	int i;
	char *ptr = img;
	for( i = 0; i < numOfObjects; i++ )	{
		Msg *msg = (Msg*) ptr;		
		switch( msg->code )	{
		case GRAPH:
			DrawGraph( (Graph*)ptr ); 
			break;
		case LINE:
			DrawLine( (Gline*)ptr ); 
			break;
		case TEXT:
			DrawText( (Gtext*)ptr ); 
			break;
		case MARKER:
			DrawMarker( (Gmarker*)ptr ); 
			break;
		case FILL:
			DrawFill( (Gfill*)ptr ); 
			break;
		case BAR:
			DrawBar( (Gbar*)ptr ); 
			break;
		}
		ptr += msg->size;
	}
}

//-----------------------------------------------------
// Redraw callback
//-----------------------------------------------------
void Redraw( PtWidget_t *widget, PhTile_t *damage )  {
	char fName[30];
	PtBasicWidgetCanvas( widget, &rect );
	marginX = rect.lr.x / 12;
	marginY = rect.lr.y / 12;
	mx = (rect.lr.x - 2 * marginX ) / XMAX;
	my = (rect.lr.y - 2 * marginY ) / YMAX ;
	sprintf( fName, "swiss%02db", rect.lr.x / 35 );
	PgSetFont( fName );		
	PgSetStrokeColor( pColor[0] );
	PgSetFillColor( pColor[0] );			// clear the screen
	PgDrawRect( &rect, Pg_DRAW_FILL|Pg_DRAW_STROKE);
	DrawObjects();
}

//-----------------------------------------------------
// If waitForMouse set, send mouse position to 'euler'
//-----------------------------------------------------
int MouseCallback( PtWidget_t *w, ApInfo_t *a, PtCallbackInfo_t *cbinfo )
{
	if( Ph_EV_RELEASE_REAL == cbinfo->event->subtype && waitForMouse)
	{
		PhRect_t *rect;
		float mouse[2];
		rect = PhGetRects( cbinfo->event );
		mouse[0] = (rect->ul.x - marginX) / mx;
		mouse[1] = (rect->ul.y - marginY) / my;
		write( STDOUT_FILENO, mouse, sizeof(mouse) );
		waitForMouse = 0;
	}
	return Pt_CONTINUE;
}

int ResizeCallback( PtWidget_t *widget, ApInfo_t *a, PtCallbackInfo_t *c )
{
	PtArg_t	args;
	PhDim_t dim;
	PtWidgetDim( widget, &dim );
	PtSetArg( &args, Pt_ARG_DIM, &dim, 0 );
	PtSetResources( plotw, 1, &args );
	return Pt_CONTINUE;
}

PtWidget_t* CreateRawWidget( PtWidget_t *window )
{
    PtArg_t args[4];
	PhDim_t dim = {400,300};
    PtRawCallback_t mousecall;
	mousecall.event_f = &MouseCallback;
	mousecall.event_mask = Ph_EV_BUT_RELEASE;
    PtSetArg(&args[0], Pt_ARG_DIM, &dim , 0); 
    PtSetArg(&args[1], Pt_ARG_BORDER_WIDTH, 0 , 0); 
    PtSetArg(&args[2], Pt_ARG_RAW_DRAW_F, &Redraw , 0); 
    PtSetArg(&args[3], Pt_CB_RAW, &mousecall, 0); 
    return PtCreateWidget(PtRaw, window, 4, args);
}

//------------------------------------------
// Add some memory
//------------------------------------------
static void CheckSpace( int size )
{
	static int bufSize = 0;
	while( size > bufSize ) {
		char* newbuf = realloc( img, bufSize + BLOCK );
		if( newbuf == 0 ) {
			char txt[40];
			free( img );
			sprintf( txt, "buffer size %d", bufSize );
			perror( txt );
			exit(1);
		}
		img = newbuf;
		bufSize += BLOCK; 
	}
}

//-----------------------------------------------------------
// Get function parameters from 'euler'
// Inputs
//	 Request for mouse position.
//   Clear screen
//	 Draw object
//-----------------------------------------------------------
int GetData( int fd, void *buf, unsigned mode )
{
	static int imgSize = 0;
	static int msgPtr = 0;
	int redraw = 0;			// Set, if needs redraw 
	while( 1 ) {
		int count;
		Msg *msg;
		int msgSize = 0;
		if( msgPtr == 0 ) {  // start of message
			count = read( STDIN_FILENO, &msgSize,
				sizeof( msgSize ) );
			if( count < 0)				// no messages
				break;
			msgPtr += count;
		}
		CheckSpace( PIPE_BUF + imgSize + msgPtr );
		count = read( STDIN_FILENO, &img[ imgSize + msgPtr],
			 min( msgSize - msgPtr, PIPE_BUF) );
		if(count > 0)
			msgPtr += count;
		if( msgSize > msgPtr ) 
			continue;
		msg = (Msg*)&img[imgSize];
		msg->size = msgSize;

		switch( msg->code ) {
		case POSITION:					// wait for mouse click
			waitForMouse = 1;
			break;
		case CLEAR:						// clear window
			imgSize = 0;	
			numOfObjects = 0;
			redraw = 1;
			break;
		default:						// all others are graphic objects
			imgSize += msgSize;
			numOfObjects ++;
			redraw = 1;
			break;
		}
		msgPtr = 0;						// clear msg buffer
	}
	if( redraw ) {
		PtDamageExtent( plotw, &rect );
	}
	return Pt_CONTINUE;
}





